/*
Created by: X
http://www.createdbyx.com/

SLPathfindaz is licenced under the Creative Commons Attribution-NonCommercial-ShareAlike 2.5 Canada
http://creativecommons.org/licenses/by-nc-sa/2.5/ca/

Special thanks to SwampThingTom from which the code in this library was originally written.
 
SwampThingTom@bayougames.com
http://www.bayougames.com/
Link to the Pathfinding Sample Using A* article and sample xna project
http://swampthingtom.blogspot.com/2007/07/pathfinding-sample-using.html
Related A* pathfinding article
http://www.policyalmanac.org/games/aStarTutorial.htm

Original Copyright header can be found in tehe OriginalLicenseHeader.txt file
    */


//
// Use the following conditional compilation symbol to allow the start and
// destination tiles to be equal.  In general, I think it's better to have the
// caller perform that optimization and never try to find a path from the
// current tile to the same tile!  But this option allows for the use of this
// algorithm even when that can't be guaranteed.
//
// #define ALLOW_EQUAL_START_STOP

using System.Collections.Generic;
using System.Diagnostics;
using Microsoft.Xna.Framework;

namespace SLPathfindaz
{
    /// <summary>
    /// Class used to find the shortest path between two map tiles.
    /// 
    /// The FindPath method implements the A* graph search algorithm to create
    /// a list of waypoints defining the most efficient path.
    /// 
    /// The GetNextWaypoint method is used to traverse the current path.
    /// </summary>
    public class PathFinder
    {
        #region Adjacent Nodes

        /// <summary>
        /// Defines map offsets and movement cost for nodes adjacent to 
        /// another node.
        /// </summary>
        struct AdjacentNode
        {
            /// <summary>
            /// The relative position of the adjacent node.
            /// </summary>
            public Point offset;

            /// <summary>
            /// The cost of moving to the adjacent node.
            /// </summary>
            public int cost;

            /// <summary>
            /// Creates an AdjacentNode.
            /// </summary>
            /// <param name="offset">The relative position of the adjacent node</param>
            /// <param name="cost">The cost of moving to the adjacent node</param>
            public AdjacentNode(Point offset, int cost)
            {
                this.offset = offset;
                this.cost = cost;
            }
        }

        /// <summary>
        /// Set of nodes adjacent to any given node.
        /// The cost of moving along a diagonal is defined as 1.4 times the 
        /// cost of moving horizontally or vertically.
        /// </summary>
        static readonly AdjacentNode[] AdjacentNodes =
            new AdjacentNode[] 
            { 
                new AdjacentNode(new Point(-1, -1), 14), 
                new AdjacentNode(new Point( 0, -1), 10), 
                new AdjacentNode(new Point( 1, -1), 14),
                new AdjacentNode(new Point(-1,  0), 10),
                new AdjacentNode(new Point( 1,  0), 10),
                new AdjacentNode(new Point(-1,  1), 14),
                new AdjacentNode(new Point( 0,  1), 10),
                new AdjacentNode(new Point( 1,  1), 14)
            };

        #endregion

        #region Private Members

        /// <summary>
        /// The map used for pathfinding.
        /// </summary>
        IMap gameMap;

        /// <summary>
        /// Collection of all of the nodes used for pathfinding.
        /// </summary>
        PathNode[,] pathMap;

        /// <summary>
        /// List of nodes making up a particular path.
        /// </summary>
        Stack<PathNode> path;

        #endregion

        #region Public Properties

        /// <summary>
        /// The map used for pathfinding.
        /// </summary>
        public IMap GameMap
        {
            get { return gameMap; }

            set
            {
                gameMap = value;
                Debug.Assert((gameMap.Width == pathMap.GetLength(0)) &&
                             (gameMap.Height == pathMap.GetLength(1)));
            }
        }

        #endregion

        #region Construction

        /// <summary>
        /// Gets or sets a open list of path nodes.
        /// </summary>
        public OpenList OpenList { get; set; }

        /// <summary>
        /// Creates a new PathFinder object.
        /// </summary>
        public PathFinder(IMap gameMap)
        {
            // PathNode.CreateOpenList(gameMap.Width, gameMap.Height);
            this.OpenList=new OpenList();
            path = new Stack<PathNode>(gameMap.Width * gameMap.Height);
            pathMap = new PathNode[gameMap.Width, gameMap.Height];

            for (int x = 0; x < gameMap.Width; x++)
            {
                for (int y = 0; y < gameMap.Height; y++)
                {
                    pathMap[x, y] = new PathNode(this.OpenList, x, y);
                }
            }

            this.gameMap = gameMap;
        }

        #endregion

        #region Find Path

        /// <summary>
        /// Finds a path from the start to the destination.
        /// </summary>
        /// <param name="start">Starting map tile</param>
        /// <param name="destination">Destination map tile</param>
        public void FindPath(Point start, Point destination)
        {
            this.FindPath(start.X, start.Y, destination.X, destination.Y);
        }


        /// <summary>
        /// Finds a path from the start to the destination.
        /// </summary>
        public void FindPath(int startX, int startY, int endX, int endY)
        {
#if !ALLOW_EQUAL_START_STOP
            //Debug.Assert(!start.Equals(destination), "Start and end points must be different.");
            Debug.Assert(!(startX == endX && startY == endY), "Start and end points must be different.");
#endif

            Debug.Assert((startX >= 0) && (startX <= gameMap.Width) &&
                         (startY >= 0) && (startY <= gameMap.Height) &&
                         (endX >= 0) && (endX <= gameMap.Width) &&
                         (endY >= 0) && (endY <= gameMap.Height));

            //
            // Reset current path, open list, and status of every node.
            //
            Reset();

            //
            // Start at the beginning.
            //
            pathMap[startX, startY].AddToOpenList();

            PathNode node;

#if ALLOW_EQUAL_START_STOP
            if (start.Equals(destination))
            {
                node = PathNode.GetBestOpenNode();
                path.Push(node);
                return;
            }
#endif

            do
            {
                //
                // Get the node that we think is closest to the destination.
                //
                //node = PathNode.GetBestOpenNode();
                node =this.OpenList.GetBestOpenNode();

                if (node == null)
                {
                    // No path is found, so stay at start location.
                    Debug.WriteLine("Pathfinder.FindPath: No path was found!");
                    pathMap[startX, startY].AddToOpenList();
                    //node = PathNode.GetBestOpenNode();
                    node = OpenList.GetBestOpenNode();
                    path.Push(node);
                    return;
                }

                //
                // Have we found our destination?
                //
                //if (!node.Tile.Equals(destination))
                if (!(node.Tile.X == endX && node.Tile.Y == endY))
                {
                    //
                    // Search each adjacent node.
                    //
                    for (int index = 0; index < AdjacentNodes.Length; index++)
                    {
                        Point adjacent;
                        adjacent.X = node.Tile.X + AdjacentNodes[index].offset.X;
                        adjacent.Y = node.Tile.Y + AdjacentNodes[index].offset.Y;

                        if ((adjacent.X >= 0) && (adjacent.X < gameMap.Width) &&
                            (adjacent.Y >= 0) && (adjacent.Y < gameMap.Height))
                        {
                            PathNode newNode = pathMap[adjacent.X, adjacent.Y];

                            if ((!newNode.IsClosed) &&
                                (gameMap.IsTileClear(adjacent)))
                            {
                                int costFromStart =
                                    node.CostFromStart +
                                    AdjacentNodes[index].cost;

                                if (!newNode.IsOpen)
                                {
                                    //
                                    // Add the node to the open list.
                                    //
                                    int tmpX = endX - adjacent.X;
                                    int tmpY = endY - adjacent.Y;
                                    int distanceFromDestination =
                                        10 * ((tmpX >= 0 ? tmpX : -tmpX) +
                                              (tmpY >= 0 ? tmpY : -tmpY));
                                    //int distanceFromDestination =
                                    //    10 * (Math.Abs(endX - adjacent.X) +
                                    //          Math.Abs(endY - adjacent.Y));

                                    newNode.AddToOpenList(
                                        node,
                                        costFromStart,
                                        distanceFromDestination);
                                }
                                else if (costFromStart < newNode.CostFromStart)
                                {
                                    //
                                    // This path is faster than the previous one
                                    // found for this node, so update it.
                                    //
                                    newNode.UpdateOpenNode(node, costFromStart);
                                }
                            }
                        }
                    }
                }
            } while (!(node.Tile.X == endX && node.Tile.Y == endY));
            //} while (!node.Tile.Equals(destination));

            //
            // Create path list.  Note that the way we are doing this
            // assumes we will always have at least two nodes in our path.
            // i.e., the start node and end node must not be equal.
            //
            do
            {
                path.Push(node);
                node = node.Parent;
            } while (node.Parent != null);
        }

        /// <summary>
        /// Reset all data structures used for pathfinding.
        /// </summary>
        void Reset()
        {
            //PathNode.ClearOpenList();
            this.OpenList.Clear();
            path.Clear();
            foreach (PathNode node in pathMap)
            {
                node.Reset();
            }
        }

        #endregion

        #region Traverse Path

        /// <summary>
        /// Determine whether the destination has been reached.
        /// </summary>
        /// <returns>True if there are no more waypoints in the current path</returns>
        public bool Complete
        {
            get { return (path.Count == 0); }
        }

        /// <summary>
        /// Gets the next waypoint from the path.
        /// It is the caller's responsibility to verify that there are more 
        /// waypoints (i.e., the Complete method returns false).
        /// </summary>
        /// <returns>Map tile of next waypoint</returns>
        public Point GetNextWaypoint()
        {
            return path.Pop().Tile;
        }

        #endregion
    }
}
