/*
Created by: X
http://www.createdbyx.com/

SLPathfindaz is licenced under the Creative Commons Attribution-NonCommercial-ShareAlike 2.5 Canada
http://creativecommons.org/licenses/by-nc-sa/2.5/ca/

Special thanks to SwampThingTom from which the code in this library was originally written.
 
SwampThingTom@bayougames.com
http://www.bayougames.com/
Link to the Pathfinding Sample Using A* article and sample xna project
http://swampthingtom.blogspot.com/2007/07/pathfinding-sample-using.html
Related A* pathfinding article
http://www.policyalmanac.org/games/aStarTutorial.htm

Original Copyright header can be found in tehe OriginalLicenseHeader.txt file
    */


using System;
using Microsoft.Xna.Framework;

namespace SLPathfindaz
{
    /// <summary>
    /// Status of node
    /// </summary>
    public enum NodeStatus
    {
        /// <summary>
        /// Node hasn't been examined.
        /// </summary>
        Unknown,

        /// <summary>
        /// Node is on list of possibilities.
        /// </summary>
        Open,

        /// <summary>
        /// Node has been traversed.
        /// </summary>
        Closed
    }

    /// <summary>
    /// Maintains status for a node in the path search tree.
    /// Each node represents a possible waypoint along the path.
    /// 
    /// The A* algorithm works by maintaining the status of every possible 
    /// node (tile) on a map.  The possible statuses are:
    /// 
    /// * Unknown
    ///   The node has not been examined.
    ///   This is the initial status for every node.
    /// 
    /// * Open
    ///   The node is being considered as a possible waypoint for the path.
    /// 
    /// * Closed
    ///   The node has been determined to not fall along the most efficient 
    ///   path.
    /// 
    /// For Open nodes, the cost of moving along the path containing this node
    /// is also maintained.  The path containing the set of nodes with the 
    /// least total cost is the one we want.
    /// 
    /// This class no longer defines a static Open List. The new implementation
    /// adds a OpenList property to the PathFinder class. When the pathfinder class
    /// creates PathNode objects the reference to the OpenList is passed to it.
    /// This allows for sharing of open lists between instances of pathFinder objects
    /// and should ease some of the threading issues. 
    /// </summary>
    public class PathNode : IComparable<PathNode>, IEquatable<PathNode>
    {

        #region Private Members

        /// <summary>
        /// Used to store the tile location
        /// </summary>
        Point tile;
        /// <summary>
        /// used to store this nodes parent
        /// </summary>
        PathNode parent;
        /// <summary>
        /// used to store the status of this node
        /// </summary>
        NodeStatus status;
        /// <summary>
        /// Stores total cost of moving through this node.
        /// </summary>
        int totalCost;
        /// <summary>
        /// Stores actual cost of moving to this node from the starting node.
        /// </summary>
        int costFromStart;
        /// <summary>
        ///stores estimated cost of moving from this node to the ending node.
        /// </summary>
        int estimatedCostToDestination;
        /// <summary>
        /// Used to store a list of nodes
        /// </summary>
        private OpenList openList;

        #endregion

        #region Properties

        /// <summary>
        /// Gets or sets the status of this node.
        /// </summary>
        public NodeStatus Status
        {
            get { return this.status; }
            set { this.status = value; }
        }

        /// <summary>
        /// Map tile represented by this node.
        /// </summary>
        public Point Tile
        {
            get { return tile; }
        }

        /// <summary>
        /// The node that led to this node.
        /// </summary>
        public PathNode Parent
        {
            get { return parent; }
        }

        /// <summary>
        /// Total cost of moving through this node.
        /// </summary>
        public int TotalCost
        {
            get { return totalCost; }
        }

        /// <summary>
        /// Actual cost of moving to this node from the starting node.
        /// </summary>
        public int CostFromStart
        {
            get { return costFromStart; }
        }

        /// <summary>
        /// Estimated cost of moving from this node to the ending node.
        /// </summary>
        public int EstimatedCostToDestination
        {
            get { return estimatedCostToDestination; }
        }

        /// <summary>
        /// True if the node is on the Open List.
        /// </summary>
        public bool IsOpen
        {
            get { return status == NodeStatus.Open; }
        }

        /// <summary>
        /// True if the node has been Closed.
        /// </summary>
        public bool IsClosed
        {
            get { return status == NodeStatus.Closed; }
        }

        #endregion

        #region Construction

        /// <summary>
        /// Create a new Node corresponding to the given map tile.
        /// </summary>
        ///<param name="list">A reference to an open list.</param>
        /// <param name="position">Map tile represented by this node</param>
        public PathNode(OpenList list, Point position)
            : this(list, position.X, position.Y)
        {

        }

        ///<summary>
        ///Create a new Node corresponding to the given map tile.
        /// </summary>
        ///<param name="list">A reference to an open list.</param>
        ///<param name="x">X location of tile.</param>
        ///<param name="y">Y location of tile</param>
        public PathNode(OpenList list, int x, int y)
        {
            this.openList = list;
            this.tile = new Point(x, y);
            Reset();
        }

        #endregion

        #region Update Node Status

        /// <summary>
        /// Used to place the first node on the Open List.
        /// </summary>
        public void AddToOpenList()
        {
            openList.Add(this);
            status = NodeStatus.Open;
            parent = null;
            totalCost = costFromStart = estimatedCostToDestination = 0;
        }

        /// <summary>
        /// Adds a node to the Open List.
        /// </summary>
        /// <param name="parent">Node that led to this node</param>
        /// <param name="costFromStart">Cost of moving to this node from the start</param>
        /// <param name="costToDest">Estimated cost of moving from this node to the destination</param>
        public void AddToOpenList(PathNode parent, int costFromStart, int costToDest)
        {
            openList.Add(this);
            status = NodeStatus.Open;
            this.parent = parent;
            this.costFromStart = costFromStart;
            this.estimatedCostToDestination = costToDest;
            totalCost = costFromStart + costToDest;
        }

        /// <summary>
        /// Updates an open node with a new parent and cost.
        /// </summary>
        /// <param name="parent">Node that led to this node</param>
        /// <param name="costFromStart">Cost of moving to this node from the start</param>
        public void UpdateOpenNode(PathNode parent, int costFromStart)
        {
            this.parent = parent;
            this.costFromStart = costFromStart;
            totalCost = costFromStart + estimatedCostToDestination;
        }

        /// <summary>
        /// Remove node from consideration.
        /// </summary>
        public void CloseNode()
        {
            status = NodeStatus.Closed;
        }

        /// <summary>
        /// Reset the node's status to Unknown.
        /// </summary>
        public void Reset()
        {
            status = NodeStatus.Unknown;
        }

        #endregion

        #region IComparable and IEquatable

        /// <summary>
        /// Implements the IComparable interface used for 
        /// sorting the nodes by total cost.
        /// </summary>
        /// <param name="that">Node to compare with this node</param>
        /// <returns>
        /// Result of comparison.  
        ///   Less than zero indicates this is less than that.  
        ///   Equal to zero indicates that this is equal to that.
        ///   Greater than zero indicates this is greater than that.
        /// </returns>
        public int CompareTo(PathNode that)
        {
            return this.totalCost.CompareTo(that.totalCost);
        }

        /// <summary>
        /// Implements the IEquatable interface.
        /// </summary>
        /// <param name="that">Node to compare with this node</param>
        /// <returns>True if nodes represent the same map tile</returns>
        public bool Equals(PathNode that)
        {
            return this.tile.Equals(that.tile);
        }

        #endregion
    }
}
